/*
 * Key Management routines for
 * Broadcom 802.11abg Networking Device Driver
 *
 * Copyright 2005-2006, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */

#include <wlc_cfg.h>
#include <typedefs.h>
#include <bcmdefs.h>
#include <osl.h>
#include <sbutils.h>
#include <bcmutils.h>
#include <wlioctl.h>
#include <bcmwpa.h>

#include <proto/802.11.h>
#include <sbhndpio.h>
#include <sbhnddma.h>
#include <hnddma.h>
#include <d11.h>
#include <wlc_rate.h>
#include <wlc_rate_sel.h>
#include <wlc_key.h>
#include <wlc_bsscfg.h>
#include <wlc_channel.h>
#include <wlc_pub.h>
#include <wlc_bsscfg.h>
#include <wlc_pio.h>
#include <wlc.h>
#include <wlc_scb.h>
#include <wl_export.h>

#define WLC_CHECK_WEP_UPDATE(wlc, update) (update)
#define RC4_STATE_NBYTES	256	

static int wlc_key_allocate(wlc_info_t *wlc);
static void wlc_key_reset(wlc_info_t *wlc, wsec_key_t *key);
static void wlc_key_hw_reallocate(wlc_info_t *wlc, int hw_index);
static void wlc_key_hw_clear_all(wlc_info_t *wlc);
static void wlc_key_write_addr(wlc_info_t *wlc, int i, const struct ether_addr *ea);


/*
 * Find a free key slot in the wsec_keys array, and allocate a new key entry.
 */
static int
wlc_key_allocate(wlc_info_t *wlc)
{
	int max_keys = WLC_MAX_WSEC_KEYS(wlc);
	int i;

	/* find a free key */
	for (i = WSEC_MAX_DEFAULT_KEYS; i < max_keys; i++) {
		if (wlc->wsec_keys[i] == NULL)
			break;
	}
	if (i >= max_keys) {
		WL_WSEC(("wl%d: out of per station key slots\n", wlc->pub.unit));
		return BCME_NORESOURCE;
	}

	wlc->wsec_keys[i] = MALLOC(wlc->pub.osh, sizeof(wsec_key_t));
	if (wlc->wsec_keys[i] == NULL) {
		WL_ERROR(("wl%d: wlc_key_allocate: out of memory, malloced %d bytes\n",
		          wlc->pub.unit, MALLOCED(wlc->pub.osh)));
		return BCME_NOMEM;
	}

	return i;
}

/*
 * Insert a key.  At exit the key can be used as an rx key
 * but not tx.  wlc_set_crypto_index() needs to be called for that.
 */
int
wlc_key_insert(wlc_info_t *wlc, wlc_bsscfg_t *bsscfg, uint32 key_len,
	uint32 key_id, uint32 key_flags, uint32 key_algo,
	uint8 *key_data, struct ether_addr *key_ea,
	wsec_iv_t *initial_iv)
{
	struct scb *scb = NULL;
	struct scb *scb_otherband = NULL;
	wsec_key_t *key;
	int wsec_idx;

	WL_WSEC(("wl%d: wlc_key_insert: index %d keylen %d addr %s\n",
	         wlc->pub.unit, key_id, key_len,
	         bcm_ether_ntoa(key_ea, eabuf)));

	ASSERT(bsscfg);

	/* check length and IV index */
	if (key_len != WEP1_KEY_SIZE &&
	    key_len != WEP128_KEY_SIZE &&
	    key_len != TKIP_KEY_SIZE &&
	    key_len != AES_KEY_SIZE) {
	  WL_ERROR(("wl%d: wlc_key_insert: unsupported key size %d\n",
	            wlc->pub.unit, key_len));
	  return BCME_BADLEN;
	}

	if (key_id >= WSEC_MAX_DEFAULT_KEYS) {
		WL_ERROR(("wl%d: wlc_key_insert: illegal key index %d\n", wlc->pub.unit, key_id));
		return BCME_BADKEYIDX;
	}

	/* set a default key */
	if (ETHER_ISNULLADDR(key_ea) || ETHER_ISBCAST(key_ea)) {
		if (BSSCFG_STA(bsscfg)) {
			/* 
			 * If the bsscfg key pointer has not already been set, set it to point to
			 * the statically allocated memory here.
			 */
			if (bsscfg->bss_def_keys[key_id] == NULL) {
				bsscfg->bss_def_keys[key_id] = &wlc->wsec_def_keys[key_id];
			}
			ASSERT(bsscfg->bss_def_keys[key_id]->idx == key_id);
			wsec_idx = key_id;
		}
		else {
			if (bsscfg->bss_def_keys[key_id]) {
				wsec_idx = WSEC_KEY_INDEX(wlc, bsscfg->bss_def_keys[key_id]);
				ASSERT((wsec_idx >= 0) && (wsec_idx < WLC_MAX_WSEC_KEYS(wlc)));
				ASSERT(bsscfg->bss_def_keys[key_id] == wlc->wsec_keys[wsec_idx]);
			}
			else {
				/* Get a new key. */
				if ((wsec_idx = wlc_key_allocate(wlc)) < 0)
					return wsec_idx;

				bsscfg->bss_def_keys[key_id] = wlc->wsec_keys[wsec_idx];
			}
		}
	}

	/* set a per station key */
	else {
		if (wlc->pub.associated && wlc->band->bandunit != CHANNEL_BANDUNIT(wlc,
			wlc->pub.current_bss.channel)) {
		}

		if (!(scb = wlc_scblookup(wlc, key_ea))) {
			WL_ERROR(("wl%d: out of scbs\n", wlc->pub.unit));
			return BCME_NOTFOUND;
		}
		if (scb->key) {
			/* replace the existing scb key */
			wsec_idx = WSEC_KEY_INDEX(wlc, scb->key);
			ASSERT((wsec_idx >= WSEC_MAX_DEFAULT_KEYS) &&
			       (wsec_idx < WLC_MAX_WSEC_KEYS(wlc)));
			ASSERT(scb->key == wlc->wsec_keys[wsec_idx]);
		}
		else if (NBANDS(wlc) > 1 &&
		         (scb_otherband = wlc_scbfindband(wlc, key_ea,
		                                          OTHERBANDUNIT(wlc))) &&
		         scb_otherband->key) {
			/* use the same key for scbs with the same MAC in each band */
			wsec_idx = WSEC_KEY_INDEX(wlc, scb_otherband->key);
			ASSERT((wsec_idx >= WSEC_MAX_DEFAULT_KEYS) &&
			       (wsec_idx < WLC_MAX_WSEC_KEYS(wlc)));
			ASSERT(scb_otherband->key == wlc->wsec_keys[wsec_idx]);
			scb->key = wlc->wsec_keys[wsec_idx];
		} else {
			/* Get a new key. */
			if ((wsec_idx = wlc_key_allocate(wlc)) < 0)
				return wsec_idx;

			scb->key = wlc->wsec_keys[wsec_idx];
		}
	}
	WL_WSEC(("wlc_key_insert: using the key idx  %d, key_id is %d\n", wsec_idx, key_id));

	/* update the key */
	/* The following line can't use WSEC_KEY macro, since the len might be 0. */
	key = wlc->wsec_keys[wsec_idx];
	bzero((char*)key, sizeof(wsec_key_t));
	key->idx = (uint8)wsec_idx;
	bcopy((char*)key_data, (char*)key->data, key_len);
	key->len = (uint8)key_len;
	key->id = (uint8)key_id;
	bcopy((char*)key_ea, (char*)&key->ea, ETHER_ADDR_LEN);
	key->aes_mode = AES_MODE_NONE;

	switch (key_len) {
	case WEP1_KEY_SIZE:
		WL_WSEC(("wl%d: wlc_key_insert: WEP (40-bit)\n", wlc->pub.unit));
		key->algo = CRYPTO_ALGO_WEP1;
		key->algo_hw = WSEC_ALGO_WEP1;
		key->iv_len = DOT11_IV_LEN;
		key->icv_len = DOT11_ICV_LEN;
		break;
	case WEP128_KEY_SIZE:
		WL_WSEC(("wl%d: wlc_key_insert: WEP (128-bit)\n", wlc->pub.unit));
		key->algo = CRYPTO_ALGO_WEP128;
		key->algo_hw = WSEC_ALGO_WEP128;
		key->iv_len = DOT11_IV_LEN;
		key->icv_len = DOT11_ICV_LEN;
		break;
	case TKIP_KEY_SIZE:
		WL_WSEC(("wl%d: wlc_key_insert: TKIP\n", wlc->pub.unit));
		key->algo = CRYPTO_ALGO_TKIP;
		key->algo_hw = WSEC_ALGO_TKIP;
		key->iv_len = DOT11_IV_TKIP_LEN;
		key->icv_len = DOT11_ICV_LEN;
		break;
	case AES_KEY_SIZE:
		switch (key_algo) {
		case CRYPTO_ALGO_AES_OCB_MSDU :
		case CRYPTO_ALGO_AES_OCB_MPDU:
			WL_WSEC(("wl%d: wlc_key_insert: AES\n", wlc->pub.unit));
			key->algo = (uint8)key_algo;
			key->algo_hw = WSEC_ALGO_AES;
			key->iv_len = DOT11_IV_AES_OCB_LEN;
			key->icv_len = DOT11_ICV_AES_LEN;
			if (key->algo == CRYPTO_ALGO_AES_OCB_MSDU)
				key->aes_mode = AES_MODE_OCB_MSDU;
			else
				key->aes_mode = AES_MODE_OCB_MPDU;
			break;
		case CRYPTO_ALGO_AES_CCM:
		default:
			WL_WSEC(("wl%d: wlc_key_insert: AES\n", wlc->pub.unit));
			key->algo = CRYPTO_ALGO_AES_CCM;
			if (scb && (scb->flags & SCB_LEGACY_AES)) {
				WL_WSEC(("wl%d: wlc_key_insert: setting pairwise key for Legacy"
					" AES\n", wlc->pub.unit));
				key->algo_hw = WSEC_ALGO_AES_LEGACY;
			} else {
				/* 
				 * If this STA is talking to a legacy AP, mark the group key
				 * as legacy also.
				 */
				if (BSSCFG_STA(bsscfg)) {
					struct scb *apscb = NULL;
					if (!(apscb = wlc_scbfind(wlc,
						&wlc->pub.current_bss.BSSID))) {
						WL_ERROR(("wl%d: wlc_key_insert: scb for associated"
							" AP not found\n", wlc->pub.unit));
						return BCME_NOTFOUND;
					}
					if (apscb->flags & SCB_LEGACY_AES) {
						WL_WSEC(("wl%d: wlc_key_insert: setting group key"
							" for Legacy AES\n", wlc->pub.unit));
						key->algo_hw = WSEC_ALGO_AES_LEGACY;
					} else
						key->algo_hw = WSEC_ALGO_AES;
				} else
					key->algo_hw = WSEC_ALGO_AES;
			}
			key->iv_len = DOT11_IV_AES_CCM_LEN;
			key->icv_len = DOT11_ICV_AES_LEN;
			key->aes_mode = AES_MODE_CCM;
			break;
		}
		break;
	}

	/* set new default key */
	if (ETHER_ISNULLADDR(key_ea) || 
		( (key->algo == CRYPTO_ALGO_WEP1 || 
		  key->algo == CRYPTO_ALGO_WEP128) && ETHER_ISBCAST(key_ea))) {
		if (key_flags & WL_PRIMARY_KEY) {
			if (WSEC_BSS_DEFAULT_KEY(bsscfg)) {
				WSEC_BSS_DEFAULT_KEY(bsscfg)->flags &= ~WSEC_PRIMARY_KEY;
			}
			bsscfg->wsec_index = key_id;
			key->flags |= WSEC_PRIMARY_KEY;
		} else if ((uint32)bsscfg->wsec_index == key_id) {
			/* this key was the primary, but the key insert cleared the flag */
			bsscfg->wsec_index = -1;
		}
	}

	/* call the tkip module for the key setup */
	if (key->algo == CRYPTO_ALGO_TKIP) {
		wl_tkip_keyset(wlc->wl, key);
	}

	WL_WSEC(("wl%d: wlc_key_insert: key algo %d algo_hw %d flags %d\n", wlc->pub.unit,
		key->algo, key->algo_hw, key->flags));

	/* check for a provided IV init value */
	wlc_key_iv_init(wlc, key, initial_iv);

	wlc_key_update(wlc, bsscfg, wsec_idx);

	if (BSSCFG_STA(bsscfg) && (wsec_idx < WSEC_MAX_DEFAULT_KEYS)) {
		bool prev_psallowed = PS_ALLOWED(wlc);

		/*
		 * A default key is used as an indication of the presence of a group
		 * key, which in turn indicates whether the port is considered open
		 * or closed
		 */
		wlc->wsec_portopen = TRUE;

		/*
		 * If PS_ALLOWED was waiting on the portopen state then
		 *    initiate the PM state change
		 */
		if (!prev_psallowed && PS_ALLOWED(wlc))
			wlc_set_pmstate(wlc, TRUE);

	}

	return 0;
}

/* 
 * If the MSSID indication flag has been changed, the default config must
 * adapt its default key table to reflect the current status of Multi-SSID:
 * if Multi-SSID is enabled, the four statically allocated keys which
 * correspond to the hardware default keys cannot be used, whereas if
 * Multi-SSID is not enabled, it is better to use the hardware default keys.
 * Thus, if the mssid flag is changed, the type of keys used must also be
 * changed, and any pre-existing keys must be copied over.
 */
void
wlc_key_mssid_change(wlc_info_t *wlc, bool old_mssid, bool new_mssid)
{
	int ii;
	int index;
	int wsec_index;
	wsec_key_t *key;

	WL_WSEC(("\nwl%d: wlc_key_mssid_change: old mssid val: %d, new mssid val: %d\n",
	         wlc->pub.unit, old_mssid, new_mssid));

	if (old_mssid == new_mssid)
		return;

	wsec_index = wlc->bsscfg[0]->wsec_index;

	for (ii = 0; ii < WLC_DEFAULT_KEYS; ii++) {
		if (wlc->bsscfg[0]->bss_def_keys[ii]) {
			key = wlc->bsscfg[0]->bss_def_keys[ii];
			wlc->bsscfg[0]->bss_def_keys[ii] = NULL;
			ASSERT(key->idx >= WSEC_MAX_DEFAULT_KEYS);
			index = ii;
			wlc->bsscfg[0]->bss_def_keys[ii] = &wlc->wsec_def_keys[ii];
			bcopy((char*)key, (char*)wlc->bsscfg[0]->bss_def_keys[ii],
			      sizeof(wsec_key_t));
			wlc->bsscfg[0]->bss_def_keys[ii]->idx = (uint8)index;

			if (key->idx >= WSEC_MAX_DEFAULT_KEYS)
				wlc_key_delete(wlc, wlc->bsscfg[0], key->idx);
		}
	}

	wlc->bsscfg[0]->wsec_index = wsec_index;
}

void
wlc_key_remove_all(wlc_info_t *wlc)
{
	int i;

	for (i = 0; i < WLC_MAX_WSEC_KEYS(wlc); i++)
		wlc_key_delete(wlc, wlc->bsscfg[0], i);
}

void
wlc_key_remove(wlc_info_t *wlc, wlc_bsscfg_t *bsscfg, wl_wsec_key_t *remove)
{
	struct scb *scb;
	wsec_key_t *key = NULL;

	/* remove a default key */
	if (ETHER_ISNULLADDR(&remove->ea) || ETHER_ISBCAST(&remove->ea)) {
		if (remove->index < WSEC_MAX_DEFAULT_KEYS) {
			key = bsscfg->bss_def_keys[remove->index];
			if (key) {
				wlc_key_delete(wlc, bsscfg, key->idx);
				bsscfg->bss_def_keys[remove->index] = NULL;
			}
		}
	}
	/* remove a per station key */
	else {
		if ((scb = wlc_scbfind(wlc, &remove->ea)) && scb->key)
			wlc_key_delete(wlc, bsscfg, WSEC_KEY_INDEX(wlc, scb->key));
		else if (NBANDS(wlc) > 1 &&
		         (scb = wlc_scbfindband(wlc, &remove->ea,
		                                OTHERBANDUNIT(wlc))) &&
		         scb->key)
		  wlc_key_delete(wlc, bsscfg, WSEC_KEY_INDEX(wlc, scb->key));
	}
}

void
wlc_key_delete(wlc_info_t *wlc, wlc_bsscfg_t *bsscfg, int index)
{
	bool prev_psallowed;
	struct scb *scb;
	wsec_key_t *key;

	ASSERT(index < WSEC_MAX_KEYS);

	key = WSEC_KEY(wlc, index);

	/* check if already deleted */
	if (key == NULL) {
		WL_ERROR(("Key is Null index %d", index));
		return;
	}

	/* call the tkip module key cleanup */
	if (key->algo == CRYPTO_ALGO_TKIP) {
		WL_ERROR(("Key Algo is TKIP"));
		key->len = 0;
		wl_tkip_keyset(wlc->wl, key);
	}
	else 
		WL_ERROR(("Key Algo is not TKIP"));

	if (BSSCFG_STA(bsscfg))
	{
		/* If Pairwise key or Group key is removed sec port is closed */
		prev_psallowed = PS_ALLOWED(wlc);
		wlc->wsec_portopen = FALSE;
		if (prev_psallowed && !PS_ALLOWED(wlc))
			wlc_set_pmstate(wlc, FALSE);
	}


	/* if this is a paired key, clean up any scb pointers */
	if (!ETHER_ISNULLADDR(&key->ea)) {
		if ((scb = wlc_scbfind(wlc, &key->ea)) &&
		    scb->key) {
			ASSERT(!bcmp(key->ea.octet, scb->ea.octet, ETHER_ADDR_LEN));
			ASSERT(key == scb->key);
			scb->key = NULL;
		}
		if (NBANDS(wlc) > 1 &&
		    (scb = wlc_scbfindband(wlc, &key->ea, OTHERBANDUNIT(wlc))) &&
		    scb->key) {
			ASSERT(!bcmp(key->ea.octet, &scb->ea, ETHER_ADDR_LEN));
			ASSERT(key == scb->key);
			scb->key = NULL;
		}
	}
	else {
		/* if this was the primary tx key, update wsec_index */
		if (key->id == bsscfg->wsec_index)
			bsscfg->wsec_index = -1;
	}

	/* reset even if we deallocate to catch use after free */
	wlc_key_reset(wlc, key);

	/* free the memory block for paired keys */
	if (index >= WSEC_MAX_DEFAULT_KEYS) {
		MFREE(wlc->pub.osh, key, sizeof(wsec_key_t));
		wlc->wsec_keys[index] = NULL;
	}

	wlc_key_update(wlc, bsscfg, index);
}

/* reset the given key to initial values */
static void
wlc_key_reset(wlc_info_t *wlc, wsec_key_t *key)
{
	ASSERT(WSEC_KEY_INDEX(wlc, key) < WSEC_MAX_KEYS);

	key->len = 0;
	key->flags = 0;
	key->algo = CRYPTO_ALGO_OFF;
	key->algo_hw = WSEC_ALGO_OFF;
	bzero(key->ea.octet, ETHER_ADDR_LEN);
}

void
wlc_scb_key_delete(wlc_info_t *wlc, struct scb *scb)
{
	int key_index;
	wsec_key_t *key;
	key = wlc_scb_getkey(scb);

	if (!key)
		return;

	key_index = WSEC_KEY_INDEX(wlc, key);

	WL_WSEC(("wl%d: %s: deleting pairwise key %d for %s\n",
	         wlc->pub.unit, __FUNCTION__,
	         key_index, bcm_ether_ntoa(&scb->ea, eabuf)));
	/* note: the scb key pointer will be set to NULL by this function. */
	wlc_key_delete(wlc, scb->bsscfg, key_index);

	/* if we are freeing up a hardware key, look for a scb using a softkey
	 * to swap in to the hardware key slot
	 */
	if (key_index < WLC_MAX_WSEC_HW_KEYS(wlc) && !WLC_SW_KEYS(wlc)) {
		wlc_key_hw_reallocate(wlc, key_index);
	}
}

static void
wlc_key_hw_reallocate(wlc_info_t *wlc, int hw_index)
{
	int i;
	int soft_index;
	uint max_used;
	struct scb *scb = NULL;
	wsec_key_t *key;
	uint bandunit;

	WL_TRACE(("wl%d: %s: hw_index = %d\n",
		wlc->pub.unit, __FUNCTION__, hw_index));

	ASSERT(wlc->wsec_keys[hw_index] == NULL);

	/*
	 * Find a candidate paired soft key to move into the hw key slot
	 */
	max_used = 0;
	soft_index = -1;

	for (i = WLC_MAX_WSEC_HW_KEYS(wlc); i < WLC_MAX_WSEC_KEYS(wlc); i++) {
		key = wlc->wsec_keys[i];
		if (key == NULL)
			continue;

		/* do the scb lookup in both bands */
		bandunit = wlc->band->bandunit;

		do {
			/* look up the scb for this key */
			scb = wlc_scbfindband(wlc, &key->ea, bandunit);

			/* if this scb is the most recently used, remember the key index */
			if (scb && (max_used < scb->used)) {
				max_used = scb->used;
				soft_index = i;
			}

			if (NBANDS(wlc) > 1)
				bandunit = (bandunit == 0) ? 1 : 0;
		} while (bandunit != wlc->band->bandunit);
	}

	/* return if no soft keys in use */
	if (soft_index == -1)
		return;

	/*
	 * A taker for the hw slot has been found.
	 */
	key = wlc->wsec_keys[soft_index];

	WL_WSEC(("wl%d: %s: swapping soft key %d to hw key %d for MAC %s\n",
		wlc->pub.unit, __FUNCTION__,
		soft_index, hw_index, bcm_ether_ntoa(&key->ea, eabuf)));

	/* move the soft key to the hw slot */
	wlc->wsec_keys[hw_index] = key;
	wlc->wsec_keys[soft_index] = NULL;
	key->idx = (uint8)hw_index;

	/* now push the new key info to the hw key */
	wlc_key_update(wlc, scb->bsscfg, hw_index);

	/* clear out the soft key */
	wlc_key_update(wlc, scb->bsscfg, soft_index);
}

void
wlc_key_iv_init(wlc_info_t *wlc, wsec_key_t *key, wsec_iv_t *initial_iv)
{
	d11regs_t *regs = wlc->regs;
	uint16 r1, r2;
	int i;

	WL_TRACE(("wl%d: wlc_key_iv_init\n", wlc->pub.unit));

	if (!wlc->pub.up)
		return;

	ASSERT(key);

	switch (key->algo) {

	case CRYPTO_ALGO_TKIP:
		break;
	case CRYPTO_ALGO_WEP1:
	case CRYPTO_ALGO_WEP128:
		/* initialize transmit IV to a ~random value to minimize IV re-use
		 * at startup
		 */
		r1 = R_REG(&regs->tsf_random);
		r2 = R_REG(&regs->tsf_random);
		key->txiv.hi = ((r2 & 0xff) << 16) | (r1 & 0xff00) | (r1 & 0x00ff);
		key->txiv.lo = 0;	/* WEP only uses hi */
		/* rx IV is not used for WEP */
		bzero(key->rxiv, sizeof(key->rxiv));
		break;

	case CRYPTO_ALGO_AES_CCM:
		for (i = 0; i < WLC_NUMRXIVS; i ++) {
			if (initial_iv != NULL) {
				key->rxiv[i].hi = initial_iv->hi; 
				key->rxiv[i].lo = initial_iv->lo;
			} else {
				key->rxiv[i].hi = 0;
				key->rxiv[i].lo = 0;
			}
		}
		key->txiv.hi = key->txiv.lo = 0;
		break;

	default:
		WL_ERROR(("wl%d: wlc_key_iv_init: unsupported algorithm %d\n", wlc->pub.unit,
			key->algo));
		break;
	}

}

#define WLC_AES_EXTENDED_PACKET	(1 << 5)
#define WLC_AES_OCB_IV_MAX	((1 << 28) - 3)

void
wlc_key_iv_update(wlc_info_t *wlc, wlc_bsscfg_t *bsscfg, wsec_key_t *key, uchar *buf, bool update)
{
	int index;
	uint32 iv_tmp = 0;
	wsec_iv_t *txiv;

	ASSERT(key);

	/* group keys in WPA-NONE (IBSS only, AES and TKIP) use a global TXIV */
	if ((bsscfg->WPA_auth & WPA_AUTH_NONE) &&
	    ETHER_ISNULLADDR(&key->ea))
		txiv = &bsscfg->wpa_none_txiv;
	else
		txiv = &key->txiv;

	WL_WSEC(("wl%d: wlc_key_iv_update start: iv32 0x%08x iv16 0x%04x algo %d \n", wlc->pub.unit,
		txiv->hi, txiv->lo, key->algo));

	index = key->id;

	switch (key->algo) {

	case CRYPTO_ALGO_AES_CCM:
		if (update) {
			txiv->lo++;
			if (txiv->lo == 0)
				txiv->hi++;
		}

		/* Add in control flags and properly format */
		buf[0] = txiv->lo & 0xff;
		buf[1] = (txiv->lo >> 8) & 0xff;
		buf[2] = 0;
		buf[3] = (uint8)(((index & 3) << DOT11_KEY_INDEX_SHIFT) | WLC_AES_EXTENDED_PACKET);
		buf[4] = txiv->hi & 0xff;
		buf[5] = (txiv->hi >>  8) & 0xff;
		buf[6] = (txiv->hi >> 16) & 0xff;
		buf[7] = (txiv->hi >> 24) & 0xff;
		break;

	case CRYPTO_ALGO_AES_OCB_MPDU:
	case CRYPTO_ALGO_AES_OCB_MSDU:
		if (update) {
			if (txiv->hi >= (1 << 28) - 3) {
				txiv->hi = 0;
			} else {
				txiv->hi++;
			}
		}

		/* Deconstruct the IV */
		buf[0] = txiv->hi & 0xff;
		buf[1] = (txiv->hi >>  8) & 0xff;
		buf[2] = (txiv->hi >> 16) & 0xff;
		buf[3] = (index & 3) << DOT11_KEY_INDEX_SHIFT;
		buf[3] |= (txiv->hi >> 24) & 0xf;
		break;

	case CRYPTO_ALGO_TKIP:
		break;

	case CRYPTO_ALGO_WEP1:
	case CRYPTO_ALGO_WEP128:
		{
		/* Handle WEP */
		uint bad = 1;
		uchar x, y, z, a, b, B;

		if (WLC_CHECK_WEP_UPDATE(wlc, update)) {

			iv_tmp = txiv->hi;
			while (bad) {
				iv_tmp++;

				x = iv_tmp & 0xff;
				y = (iv_tmp >> 8) & 0xff;
				z = (iv_tmp >> 16) & 0xff;

				a = x + y;
				b = (x + y) - z;

				/* Avoid weak IVs used in several published WEP crackingtools. */
				if (a <= WEP128_KEY_SIZE)
					continue;

				if (a == 1 && b == 2)
					continue;

				if ((y == RC4_STATE_NBYTES - 1) &&
					(x > 2 && x <= WEP128_KEY_SIZE + 2))
					continue;

				if (x == 1 && (y >= 2 && y <= ((WEP128_KEY_SIZE-1)/2 + 1)))
					continue;

				bad = 0;
				B = 1;
				while (B < WEP128_KEY_SIZE) {
					if ((a == B && b == (B + 1) * 2)) {
						bad++;
						break;
					}
					B += 2;
				}
				if (bad) continue;

				B = 3;
				while (B < WEP128_KEY_SIZE/2 + 3) {
					if ((x == B) && (y == (RC4_STATE_NBYTES-1)-x)) {
						bad++;
						break;
					}
					B++;
				}
			}

			buf[0] = x;
			buf[1] = y;
			buf[2] = z;
			buf[3] = index << DOT11_KEY_INDEX_SHIFT;
			txiv->hi = iv_tmp;
		} else {
			/* Don't update - just break down in 3 bytes */
			buf[0] = txiv->hi & 0xff;
			buf[1] = (txiv->hi >> 8) & 0xff;
			buf[2] = (txiv->hi >> 16) & 0xff;
			buf[3] = index << DOT11_KEY_INDEX_SHIFT;
		}

		}
		break;
	default:
		WL_ERROR(("wl%d: wlc_key_iv_update: unsupported algorithm %d\n", wlc->pub.unit,
			key->algo));
		ASSERT(1 == 0);
		break;
	}

	WL_WSEC(("wl%d: wlc_key_iv_update end: iv32 0x%08x iv16 0x%04x\n", wlc->pub.unit, txiv->hi,
		txiv->lo));
}

wsec_key_t*
wlc_key_lookup(wlc_info_t *wlc, struct scb *scb, wlc_bsscfg_t *bsscfg, uint index, bool ismulti)
{
	wsec_key_t *key = NULL;

	WL_WSEC(("\nwl%d: wlc_key_lookup: index %d\n", wlc->pub.unit, index));

	if (scb->key && !ismulti) {
		key = scb->key;
		ASSERT(WSEC_KEY_INDEX(wlc, key) < WLC_MAX_WSEC_KEYS(wlc));
		/* ASSERT(key->index == index); */
		/* indexes should both be zero for per-path key */
		WL_WSEC(("wl%d: wlc_key_lookup: per-path key-ID %d index %d, algo %d\n",
			wlc->pub.unit,
			key->id, key->idx, key->algo));
	} else {
		ASSERT(index < WSEC_MAX_DEFAULT_KEYS);
		key = WLC_BSSCFG_WEPKEY(bsscfg, index);
		if (key) {
			WL_WSEC(("wl%d: wlc_key_lookup: valid group key at key-ID %d, algo %d\n",
				wlc->pub.unit, index, key->algo));
		} else {
			WL_WSEC(("wl%d: wlc_key_lookup: no valid group key at key-ID %d\n",
				wlc->pub.unit, index));
		}
	}

	return key;
}

/* initialize all hardware keys for the current core */
void
wlc_key_hw_init_all(wlc_info_t *wlc)
{
	uint i;
	uint hw_keys;

	WL_TRACE(("wl%d: wlc_key_hw_init_all\n", wlc->pub.unit));

	wlc_key_hw_clear_all(wlc);

	if (WLC_SW_KEYS(wlc))
		return;

	hw_keys = WLC_MAX_WSEC_HW_KEYS(wlc);

	for (i = 0; i < hw_keys; i++) {
		if (WSEC_KEY(wlc, i))
			wlc_key_hw_init(wlc, i);
	}
}

/* Clear all hardware key state in the current core.
 * This assumes that the core has been reset so all registers
 * have their reset value.
 */
static void
wlc_key_hw_clear_all(wlc_info_t *wlc)
{
	struct ether_addr addr;
	int i, hw_keys, num_addrs;
	uint16 v;

	WL_TRACE(("wl%d: wlc_key_hw_clear_all\n", wlc->pub.unit));

	hw_keys = WLC_MAX_WSEC_HW_KEYS(wlc);

	/* zero out the entire key table */
	wlc_set_shm(wlc, wlc->seckeys, 0, hw_keys * D11_MAX_KEY_SIZE);

	for (i = 0; i < hw_keys; i++) {
		/* Set the matching SECALGO table entry */
		v = WSEC_ALGO_OFF | (i << SKL_INDEX_SHIFT);
		wlc_write_shm(wlc, M_SECKINDXALGO_BLK + i*2, v);
	}

	/* clear the Rx-TA MAC addresses */
	bzero(addr.octet, ETHER_ADDR_LEN);
	WL_NONE(("**** KEY SETTING\n"));
	if (D11REV_LT(wlc->pub.corerev, 5)) {
		WL_NONE(("**** KEY SETTING SET ADDR corerev %d \n", wlc->pub.corerev));
		/* clear the 4 Rx-TA addrs in the RXE COND registers */
		wlc_set_addrmatch(wlc, RCM_WEP_TA0_OFFSET, &addr);
		wlc_set_addrmatch(wlc, RCM_WEP_TA1_OFFSET, &addr);
		wlc_set_addrmatch(wlc, RCM_WEP_TA2_OFFSET, &addr);
		wlc_set_addrmatch(wlc, RCM_WEP_TA3_OFFSET, &addr);

		/* clear the soft Rx-TA addrs in shm */
		num_addrs = WSEC_MAX_SEC_KEYS - (WSEC_MAX_DEFAULT_KEYS + WSEC_MAX_RXE_KEYS);
		wlc_set_shm(wlc, M_SECPSMRXTAMCH_BLK, 0, num_addrs * ETHER_ADDR_LEN);
		/* set the soft Rx-TA addr count */
		wlc_write_shm(wlc, M_SEC_VALNUMSOFTMCHTA, WSEC_MAX_SEC_KEYS -
			(WSEC_MAX_DEFAULT_KEYS + WSEC_MAX_RXE_KEYS));
	} else {
		WL_NONE(("**** KEY SETTING RCMTA corerev %d \n", wlc->pub.corerev));
		/* clear the address table */
		for (i = 0; i < (WSEC_MAX_RCMTA_KEYS - WSEC_MAX_DEFAULT_KEYS); i++)
			wlc_set_rcmta(wlc, i, &addr);

		/* update the rcmta_size reg in case we are running with a
		 * WSEC_MAX_RCMTA_KEYS value different than the reset value
		 */
		W_REG(&wlc->regs->rcmta_size, (WSEC_MAX_RCMTA_KEYS - WSEC_MAX_DEFAULT_KEYS));
	}

	/* clear the ucode default key flag MHF_DEFKEYVALID */
	wlc_mhf(wlc, MHF_DEFKEYVALID, 0, TRUE);
	WL_WSEC(("wl%d: wlc_key_hw_clear_all: clear MHF_DEFKEYVALID\n", wlc->pub.unit));
}

/* Updates all hw keys marked for band switch update in wlc->wsec_bsupdate */
void
wlc_key_bsupdate(wlc_info_t *wlc)
{
	int i;

	/* update and clear the update flag if set */
	for (i = 0; i < WLC_MAX_WSEC_HW_KEYS(wlc); i++) {
		if (isset(wlc->wsec_bsupdate.vec, i)) {
			clrbit(wlc->wsec_bsupdate.vec, i);
			wlc_key_hw_init(wlc, i);
		}
	}
}

/*
 * Return MHF_DEFKEYVALID flag based on soft keys and/or default WEP keys
 * existances and other factors.
 */
/*
 * ucode flag MHF_DEFKEYVALID should be SET only when WEP encryption is
 * enabled, only one SSID is in use, there are valid default WEP keys and
 * there are no soft keys. It is cleared in all other cases.
 * The following explains why.
 *
 * This flag is meaningful only to encrypted unicast frames received. When
 * processing a rx'd unicast frame the ucode tries a pairwise key match first
 * based on the frame's TA and then falls back to a default key indicated by
 * the key index field in IV field of the frame if no pairwise key match is
 * found. For multicast/broadcast frames the ucode uses the key index field
 * in the frames' IV anyway to find the key entry.
 *
 * In case of Multi-SSID, the default keys are only valid for one SSID, and
 * the ucode does not distinguish between the SSIDs when checking the flag.
 *
 * In case of software encryption/decryption when soft keys are installed
 * this flag must be cleared so that encrypted unicast frames can be sent up
 * to the driver without any decryption attempt when there is no key installed
 * in the hardware for the TA. It is true for both STA and AP.
 *
 * In case of WPA/TKIP|AES we don't support group-key-only configuration and
 * there must be a pairwise key there on STA and one pairwise key per
 * associated STA on AP therefore this flag should always be cleared on both
 * STA and AP.
 *
 * In case of mixed WPA/TKIP|AES and 802.1x/WEP this flag should be always
 * cleared on AP since there must be a pairwise key for each STA.
 *
 * In case of 802.1x/WEP it should be always cleared on AP. However it must
 * be set on STA since there is no pairwise key on STA.
 *
 * In case of mixed WPA/TKIP|AES and 802.11/WEP this flag should be always
 * cleared on AP since the current transition mode support on AP plumbs
 * pairwise key for WEP STAs too.
 *
 * In case of multiple SSID configuration (AP-STA and/or AP-MSSID), default
 * keys are invalid.
 * In any other case of 802.11/WEP this flag should always be set. It is true
 * for both STA and AP.
 */

uint32
wlc_key_defkeyflag(wlc_info_t *wlc)
{
	bool defkeys = FALSE, swkeys = FALSE;
	wsec_key_t *key;
	int j;

	/* no default key attempt for WPA */
	/* assuming AES or TKIP is enabled */
	if (wlc->cfg.WPA_auth != WPA_AUTH_DISABLED)
		return 0;

	/* set defkeys to TRUE if there are any default WEP keys */
	/* assuming WEP is enabled */
	for (j = 0; j < WSEC_MAX_DEFAULT_KEYS; j ++) {
		if ((key = WSEC_KEY(wlc, j)) &&
		    (key->algo == CRYPTO_ALGO_WEP1 || key->algo == CRYPTO_ALGO_WEP128)) {
			defkeys = TRUE;
			break;
		}
	}
	/* set swkeys to TRUE if there are any soft keys */
	for (j = WLC_MAX_WSEC_HW_KEYS(wlc); j < WLC_MAX_WSEC_KEYS(wlc); j++) {
		if (WSEC_KEY(wlc, j)) {
			swkeys = TRUE;
			break;
		}
	}

	/* set the flag only when there are default WEP keys and there are no soft keys */
	WL_WSEC(("wl%d: wlc_key_defkeyflag: default WEP keys %s, soft keys %s\n",
		wlc->pub.unit, defkeys ? "Yes" : "No", swkeys ? "Yes" : "No"));
	return (defkeys && !swkeys) ? MHF_DEFKEYVALID : 0;
}

/* Carry out any hardware actions needed after a wlc->wsec_key has been modified */
void
wlc_key_update(wlc_info_t *wlc, wlc_bsscfg_t *bsscfg, int index)
{
	wsec_key_t *key;

	WL_TRACE(("wl%d: wlc_key_update key %d\n", wlc->pub.unit, index));
	
	key = WSEC_KEY(wlc, index);
	if ((key) && (key->algo == CRYPTO_ALGO_TKIP))
		return;

	/*
	 * Clear the ucode default key valid flag MHF_DEFKEYVALID if inserting
	 * a per-STA soft key
	 */
	if (index >= WLC_MAX_WSEC_HW_KEYS(wlc) && WSEC_KEY(wlc, index)) {
		wlc_mhf(wlc, MHF_DEFKEYVALID, 0, TRUE);
		WL_WSEC(("wl%d: wlc_key_update: insert soft key %d, clear MHF_DEFKEYVALID\n",
		         wlc->pub.unit, index));
	}
	/*
	 * Need to check both soft keys and default WEP keys before
	 * making decision to clear or set this flag when deleting a soft key
	 */
	else if ((index >= WLC_MAX_WSEC_HW_KEYS(wlc) && !WSEC_KEY(wlc, index))) {
		uint32 flag = wlc_key_defkeyflag(wlc);
		wlc_mhf(wlc, MHF_DEFKEYVALID, flag, TRUE);
		WL_WSEC(("wl%d: wlc_key_update: delete soft key %d, %s MHF_DEFKEYVALID\n",
		         wlc->pub.unit, index, flag ? "set" : "clear"));
	}

	/* if this is a soft key, then the hardware needs no update */
	if (index >= WLC_MAX_WSEC_HW_KEYS(wlc))
		return;

	/* since we only update the hardware for the current core,
	 * mark the key for update on band switch if we are running on
	 * multiple cores
	 */
	if (wlc->ncores > 1)
		setbit(wlc->wsec_bsupdate.vec, index);

	/* update the hardware */
	if (wlc->pub.up)
		wlc_key_hw_init(wlc, index);
}

/* update the hardware with values to match the corresponding wlc->wsec_key */
void
wlc_key_hw_init(wlc_info_t *wlc, int index)
{
	wsec_key_t *key;
	struct ether_addr ea;
	uint offset;
	uint16 v;

	WL_TRACE(("wl%d: wlc_key_hw_init key %d\n", wlc->pub.unit, index));

	ASSERT(index < WLC_MAX_WSEC_HW_KEYS(wlc));

	ASSERT(wlc->sbclk);

	if (!WLC_SW_KEYS(wlc))
		key = WSEC_KEY(wlc, index);
	else
		key = NULL;

	/* Clear the MAC address match if this is a paired key.
	 * This protects the ucode or hardware from getting an
	 * address match while we update this key.
	 */
	if (index >= WSEC_MAX_DEFAULT_KEYS) {
		bzero(ea.octet, ETHER_ADDR_LEN);
		wlc_key_write_addr(wlc, index, &ea);
	}

	/* Set the matching SECALGO table entry (algorithm and table index) */
	if (!key)
		v = WSEC_ALGO_OFF | (index << SKL_INDEX_SHIFT);
	else
		v = key->algo_hw | (index << SKL_INDEX_SHIFT);


	wlc_write_shm(wlc, M_SECKINDXALGO_BLK + index * 2, v);

	/*
	 * Set the rx and tx default key (i < WSEC_MAX_DEFAULT_KEYS)
	 *	or set the per sta key (i >= WSEC_MAX_DEFAULT_KEYS).
	 */
	offset = index * D11_MAX_KEY_SIZE;

	if (key) {
		/* Tx default key */
		if (index < WSEC_MAX_DEFAULT_KEYS)
			wlc_copyto_shm(wlc, wlc->seckeys + offset, key->data, D11_MAX_KEY_SIZE);

		/* otherwise, write the key */
		wlc_copyto_shm(wlc, wlc->seckeys + WSEC_RXKEYOFF + offset, key->data,
				D11_MAX_KEY_SIZE);
	} else {
		/* zero out key data if no key */

		/* Tx default key */
		if (index < WSEC_MAX_DEFAULT_KEYS)
			wlc_set_shm(wlc, wlc->seckeys + offset, 0, D11_MAX_KEY_SIZE);
		/* Rx key */
		wlc_set_shm(wlc, wlc->seckeys + WSEC_RXKEYOFF + offset, 0, D11_MAX_KEY_SIZE);
	}

	/* handle per station keys */
	if (key && index >= WSEC_MAX_DEFAULT_KEYS) {
		/* update the MAC address match */
		wlc_key_write_addr(wlc, index, &key->ea);
	}

	/* update ucode flag MHF_DEFKEYVALID */
	/*
	 * Need to check both soft keys and default WEP keys before
	 * making decision to either set or clear it when updating a default key
	 */
	if (index < WSEC_MAX_DEFAULT_KEYS) {
		uint32 flag = wlc_key_defkeyflag(wlc);
		wlc_mhf(wlc, MHF_DEFKEYVALID, flag, TRUE);
		WL_WSEC(("wl%d: wlc_key_hw_init: %s default key %d, %s MHF_DEFKEYVALID\n",
		         wlc->pub.unit, key ? "insert" : "delete", index, flag ? "set" : "clear"));
	}
}

/* update the pairwise key address for a hardware key */
static void
wlc_key_write_addr(wlc_info_t *wlc, int i, const struct ether_addr *ea)
{
	uint offset;

	WL_TRACE(("wl%d: wlc_key_write_addr key %d\n", wlc->pub.unit, i));

	ASSERT(i < WLC_MAX_WSEC_HW_KEYS(wlc));
	ASSERT(i >= WSEC_MAX_DEFAULT_KEYS);

	if (D11REV_LT(wlc->pub.corerev, 5)) {
		/* set 4 RXE match registers */
		if (i < (WSEC_MAX_DEFAULT_KEYS + WSEC_MAX_RXE_KEYS)) {
			/* calculate register offset */
			offset = (i - WSEC_MAX_DEFAULT_KEYS) * ETHER_ADDR_LEN / 2;
			/* write to RXE match register */
			wlc_set_addrmatch(wlc, RCM_WEP_TA0_OFFSET + offset, ea);
		} else {
			/* set 8 PSM match registers */
			/* calculate shared memory offset */
			offset = (i - (WSEC_MAX_DEFAULT_KEYS + WSEC_MAX_RXE_KEYS)) * ETHER_ADDR_LEN;
			/* write to shared memory */
			wlc_copyto_shm(wlc, M_SECPSMRXTAMCH_BLK + offset, ea->octet,
				ETHER_ADDR_LEN);
		}
	} else {
		wlc_set_rcmta(wlc, (i - WSEC_MAX_DEFAULT_KEYS), ea);
	}
}


